#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/ingest/T0Compiler.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                11.05.2021
# Last Modified Date:  25.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.content.MetaRecord import MetaRecord
from ampel.types import ChannelId, DataPointId
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsDocIngester import AbsDocIngester
from ampel.util.collections import try_reduce
from ampel.abstract.AbsCompiler import AbsCompiler
from ampel.enum.MetaActionCode import MetaActionCode


class T0Compiler(AbsCompiler):
	"""
	Compiles updates to t0 documents arising from different t0 directives
	"""

	def __init__(self, **kwargs) -> None:
		super().__init__(**kwargs)

		# We assume only the usage of only one shaper (that is only one tracid)
		self.register: dict[
			DataPointId,
			tuple[
				DataPoint,
				set[ChannelId],
				None | int,           # trace id
				None | dict[str, Any] # meta extra
			]
		] = {}


	# Override
	def add(self, # type: ignore[override]
		dps: list[DataPoint],
		channel: ChannelId,
		trace_id: None | int,
		extra: None | dict[str, Any] = None
	) -> None:

		r = self.register
		for dp in dps:
			dpid = dp['id']
			if dpid in r:
				r[dpid][1].add(channel)
			else:
				r[dpid] = dp, {channel}, trace_id, extra


	# Override
	def commit(self, ingester: AbsDocIngester[DataPoint], now: int | float, **kwargs) -> None:
		"""
		Note that we let the ingester handle 'ts' and 'updated' values
		"""

		for dp, channel_sets, trace_id, extra in self.register.values():

			lchans = list(channel_sets)
			meta: MetaRecord = {'ts': now, 'run': self.run_id}

			if extra:
				meta.update(extra) # type: ignore[typeddict-item]

			meta['activity'] = [
				{
					'action': MetaActionCode.ADD_CHANNEL,
					'channel': try_reduce(lchans)
				}
			]
			meta['traceid'] = {'shaper': trace_id}

			if self._tag or dp.get('tag'):
				dp['tag'] = [*(self._tag or []), *(dp.get('tag') or [])]
				meta['activity'].append(self._ingest_tag_activity) # type: ignore

			if self.origin and 'origin' not in dp:
				dp['origin'] = self.origin

			dp['channel'] = lchans
			if 'meta' in dp and isinstance(dp['meta'], list):
				dp['meta'].append(meta)
			else:
				dp['meta'] = [meta]

			ingester.ingest(dp)

		self.register.clear()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/purge/PurgeModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                18.06.2020
# Last Modified Date:  18.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.purge.PurgeContentModel import PurgeContentModel
from ampel.model.purge.PurgeLogsModel import PurgeLogsModel


class PurgeModel(AmpelBaseModel):
	content: PurgeContentModel
	logs: PurgeLogsModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsT3Selector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.12.2019
# Last Modified Date:  17.02.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import ClassVar
from collections.abc import Iterable
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.core.ContextUnit import ContextUnit


class AbsT3Selector(AmpelABC, ContextUnit, abstract=True):

	#: field used to identify stocks
	field_name: ClassVar[str] = "stock"

	@abstractmethod
	def fetch(self) -> None | Iterable:
		""" Get selected stock ids """



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/secret/PotemkinSecretProvider.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                14.08.2020
# Last Modified Date:  20.06.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import get_args, get_origin
from ampel.abstract.AbsSecretProvider import AbsSecretProvider
from ampel.secret.Secret import Secret
from ampel.secret.NamedSecret import NamedSecret


class PotemkinSecretProvider(AbsSecretProvider):

    def tell(self, arg: Secret, value_type: type) -> bool:

        if get_origin(value_type) is tuple:
            value = tuple(t() for t in get_args(value_type))
        else:
            value = value_type() # type: ignore[assignment]
        arg.set(value)
        return True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/ProcessCommand.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  14.08.2022
# Last Modified By:    jvs

import signal, traceback, yaml
from argparse import ArgumentParser
from time import time
from typing import Any, Optional, Sequence, Union

from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.model.ChannelModel import ChannelModel
from ampel.model.UnitModel import UnitModel
from ampel.util.freeze import recursive_freeze

def _handle_traceback(signal, frame):
    print(traceback.print_stack(frame))

class ProcessCommand(AbsCoreCommand):
    """
    Runs a single process from a task definition (YAML file)

    A process is a single step of a Job (see JobCommand), with all templates resolved.
    """

    def __init__(self):
        self.parser = None

    @staticmethod
    def get_sub_ops() -> None | list[str]:
        return None

    # Mandatory implementation
    def get_parser(
        self, sub_op: Optional[str] = None
    ) -> Union[ArgumentParser, AmpelArgumentParser]:

        if self.parser:
            return self.parser

        parser = AmpelArgumentParser("process")
        parser.set_help_descr(
            {
                "debug": "enable traceback printing",
                "log-profile": "logging profile to use",
                "config": "path to an ampel config file (yaml/json)",
                "schema": "path to YAML job file",
                "name": "process name",
                "secrets": "path to a secret store; either SOPS YAML or mounted k8s Secret directory",
                "db": "database to use",
                "channel": "path to YAML channel file",
                "alias": "path to YAML alias file",
            }
        )

        parser.req("config")
        parser.req("schema")
        parser.req("name")
        parser.req("db", type=str)

        parser.opt("channel")
        parser.opt("alias")
        parser.opt("log-profile", default="prod")
        parser.opt("debug", default=False, action="store_true")
        parser.opt("secrets", type=str)

        # Example
        parser.example("process -config ampel_conf.yaml schema task_file.yaml -db processing -name taskytask")
        return parser


    def _get_context(self,
        args: dict[str, Any],
        unknown_args: Sequence[str],
        logger: AmpelLogger,
    ) -> DevAmpelContext:

        # DevAmpelContext hashes automatically confid from potential IngestDirectives
        ctx = super().get_context(
            args,
            unknown_args,
            logger,
            freeze_config=False,
            ContextClass=DevAmpelContext,
            purge_db=False,
            db_prefix=args["db"],
            require_existing_db=False,
            one_db=True,
        )

        config_dict = ctx.config._config

        # load channels if provided
        if args["channel"]:
            with open(args["channel"]) as f:
                for c in yaml.safe_load(f):
                    chan = ChannelModel(**c)
                    logger.info(f"Registering job channel '{chan.channel}'")
                    dict.__setitem__(config_dict["channel"], str(chan.channel), c)
        
        # load custom aliases if provided
        if args["alias"]:
            with open(args["alias"]) as f:
                for k, v in yaml.safe_load(f).items():
                    if k not in ("t0", "t1", "t2", "t3"):
                        raise ValueError(f"Unrecognized alias: {k}")
                    if "alias" not in config_dict:
                        dict.__setitem__(config_dict, "alias", {})
                    for kk, vv in v.items():
                        logger.info(f"Registering job alias '{kk}'")
                        if k not in config_dict["alias"]:
                            dict.__setitem__(config_dict["alias"], k, {})
                        dict.__setitem__(config_dict["alias"][k], kk, vv)
        
        ctx.config._config = recursive_freeze(config_dict)
        
        return ctx


    def run(self,
        args: dict[str, Any],
        unknown_args: Sequence[str],
        sub_op: Optional[str] = None,
    ) -> None:

        if args["debug"]:
            signal.signal(signal.SIGUSR1, _handle_traceback)

        start_time = time()
        logger = AmpelLogger.get_logger(base_flag=LogFlag.MANUAL_RUN)

        logger.info(f"Running task {args['name']}")

        with open(args["schema"], "r") as f:
            unit_model = UnitModel(**yaml.safe_load(f))
        # always raise exceptions
        unit_model.override = (unit_model.override or {}) | {"raise_exc": True}

        ctx = self._get_context(
            args,
            unknown_args,
            logger,
        )

        proc = ctx.loader.new_context_unit(
            model=unit_model,
            context=ctx,
            process_name=args["name"],
            sub_type=AbsEventUnit,
            base_log_flag=LogFlag.MANUAL_RUN,
            log_profile=args["log_profile"],
        )
        x = proc.run()
        logger.info(f"{unit_model.unit} return value: {x}")

        dm = divmod(time() - start_time, 60)
        logger.info(
            "Task processing done. Time required: %s minutes %s seconds\n"
            % (round(dm[0]), round(dm[1]))
        )
        logger.flush()
        # signal.signal(signal.SIGALRM, _handle_traceback)
        # signal.alarm(1)
        print("ProcessCommand.run() done; returning")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/T3PlainUnitExecutor.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                12.12.2021
# Last Modified Date:  13.07.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from typing import Annotated
from collections.abc import Generator

from ampel.types import ChannelId
from ampel.view.T3Store import T3Store
from ampel.view.T3DocView import T3DocView
from ampel.abstract.AbsT3ControlUnit import AbsT3ControlUnit
from ampel.abstract.AbsT3PlainUnit import AbsT3PlainUnit
from ampel.base.AmpelUnit import AmpelUnit
from ampel.t3.T3DocBuilder import T3DocBuilder
from ampel.content.T3Document import T3Document
from ampel.model.UnitModel import UnitModel
from ampel.util.hash import build_unsafe_dict_id
from ampel.util.mappings import dictify
from ampel.util.serialize import walk_and_encode


class SkippableUnitModel(UnitModel):
	"""
	Whether unit result from DB should be used if avail.
	If compatible cached result is found,
	computation of the underlying T3 unit will be skipped

	Potentialy save hash of dependent unit result (body) in meta record (extra.unit#confid).
	"""
	cache: None | UnitModel = None


class T3PlainUnitExecutor(AbsT3ControlUnit, T3DocBuilder):

	# Require single channel for now (super classes allow multi-channel)
	channel: None | ChannelId = None

	target: Annotated[SkippableUnitModel, AbsT3PlainUnit]

	def process(self, t3s: T3Store) -> None | Generator[T3Document, None, None]:

		t3_unit = self.context.loader.new_safe_logical_unit(
			UnitModel(unit=self.target.unit, config=self.target.config),
			unit_type = AbsT3PlainUnit,
			logger = self.logger,
			_chan = self.channel
		)

		if self.target.cache:

			col = self.context.db.get_collection('t3')
			h = self.get_target_cache_config()

			if view := t3s.get_view(unit=self.target.cache.unit, config=h):
				if (body := view.get_body()) and isinstance(body, dict):
					matchd = {
						'unit': self.target.unit,
						'confid': build_unsafe_dict_id(t3_unit._get_trace_content()),
						f'meta.extra.{self.target.cache.unit}#{h}': build_unsafe_dict_id(
							walk_and_encode(dictify(body), destructive=False)
						)
					}
					if (d := next(iter(col.find(matchd)), None)):
						self.logger.info("Using cached result", extra=matchd)
						t3s.add_view(
							T3DocView.of(d, self.context.config)
						)
						return None

		self.logger.info("Running T3 unit", extra={'unit': self.target.unit})
		ts = time()
		ret = t3_unit.process(t3s)
		self.flush(t3_unit)

		if ret and (x := self.handle_t3_result(t3_unit, ret, t3s, None, ts)):

			if self.target.cache and isinstance(x['body'], dict):

				if 'extra' not in x['meta']:
					x['meta']['extra'] = {}

				h = self.get_target_cache_config()
				if view := t3s.get_view(unit=self.target.cache.unit, config=h):
					x['meta']['extra'][f"{self.target.cache.unit}#{h}"] = build_unsafe_dict_id(
						walk_and_encode(dictify(view.get_body()), destructive=False)
					)

			yield x


	def get_target_cache_config(self) -> int:

		if self.target.cache:
			if isinstance(self.target.cache.config, dict):
				vc = self.context \
					.loader.get_class_by_name(self.target.cache.unit, unit_type=AmpelUnit) \
					.validate(self.target.cache.config)

				return build_unsafe_dict_id(dictify(vc))
			elif isinstance(self.target.cache.config, int):
				return self.target.cache.config
			else:
				print(type(self.target.cache.config))

		raise ValueError("Bad config")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/JobCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                15.03.2021
# Last Modified Date:  22.10.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import tarfile, tempfile, ujson, yaml, io, os, signal, sys, \
	subprocess, platform, shutil, filecmp, psutil
from time import time, sleep
from multiprocessing import Queue, Process
from argparse import ArgumentParser
from importlib import import_module
from typing import Any
from collections.abc import Sequence
from urllib.request import urlretrieve
from ampel.abstract.AbsEventUnit import AbsEventUnit
from ampel.abstract.AbsProcessorTemplate import AbsProcessorTemplate
from ampel.model.UnitModel import UnitModel
from ampel.core.EventHandler import EventHandler
from ampel.dev.DevAmpelContext import DevAmpelContext
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.util.freeze import recursive_freeze
from ampel.util.hash import build_unsafe_dict_id
from ampel.util.distrib import get_dist_names
from ampel.util.collections import try_reduce
from ampel.cli.config import get_user_data_config_path
from ampel.cli.utils import _maybe_int
from ampel.cli.AbsCoreCommand import AbsCoreCommand
from ampel.cli.MaybeIntAction import MaybeIntAction
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.model.job.JobModel import JobModel
from ampel.model.job.InputArtifact import InputArtifact
from ampel.model.job.TaskUnitModel import TaskUnitModel
from ampel.model.job.TemplateUnitModel import TemplateUnitModel
from ampel.util.pretty import out_stack, get_time_delta
from ampel.util.debug import MockPool


class JobCommand(AbsCoreCommand):
	"""
	Processes ampel "jobs" (yaml files).
	A job:
	- is a sharable analysis shema
	- contains configurations of processor units to be run sequentially
	- should also contain everything required to run the underlying processors, that is:
	  channels or aliases definitions and custom resources potentially required by processors.
	- requires access to an ampel config file since the database, unit and resource definitions
	  defined therein are necessary (might be improved in the future)
	"""

	def __init__(self):
		self.parser = None
		try:
			import sys, IPython
			sys.breakpointhook = IPython.embed
		except Exception:
			pass

	@staticmethod
	def get_sub_ops() -> None | list[str]:
		return None

	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if self.parser:
			return self.parser

		parser = AmpelArgumentParser('job')
		parser.args_not_required = True
		parser.set_help_descr({
			'debug': 'Debug',
			#'verbose': 'increases verbosity',
			'config': 'path to an ampel config file (yaml/json)',
			'schema': 'path to YAML job file (multiple files will be aggregated)',
			'secrets': 'path to a YAML secrets store in sops format',
			'keep-db': 'do not reset databases even if so requested by job file',
			'reset-db': 'reset databases even if not requested by job file',
			'no-agg': 'enables display of matplotlib plots via plt.show() for debugging',
			'no-breakpoint': 'ignore pdb breakpoints',
			'no-mp': 'deactivates multiprocessing by monkey patching multiprocessing.Pool with ampel.util.debug.MockPool for debugging',
			'interactive': 'you will be asked for each task whether it should be run or skipped\n' + \
				'and - if applicable - if the db should be reset. Option -task will be ignored.',
			'edit': 'edit job schema(s) before execution. Customize env variable EDITOR if need be. Default: "raw":\n' +
					'"raw": edit raw schema file(s) before parsing\n' +
					'"parsed": edit (possibly aggregated) schema after parsing by the yaml interpreter\n' +
					'"model": edit job model after templating (if applicable). Only task-related changes will be taken into account',
			'keep-edits': 'do not remove edited job schemas from temp dir',
			'fzf': 'choose schema file using fzf (linux/max only, fzf command line utility must be installed)',
			'task': 'only execute task(s) with provided index(es) [starting with 0].\n' +
					'Value "last" is supported. Ignored if -interactive is used',
			'show-plots': 'show plots created by job (requires ampel-plot-cli. Use "export AMPEL_PLOT_DPI=300" to increase png quality)',
			'show-plots-cmd': 'show command required to show plots created by job (requires ampel-plot-cli)',
			'wait-pid': 'wait until PID completition before starting the current job'
		})

		parser.req('config', type=str)
		parser.opt('schema', nargs='+')
		parser.opt('task', action=MaybeIntAction, nargs='+')
		parser.opt('interactive', action='store_true')
		parser.opt('debug', action='store_true')
		parser.opt('edit', nargs='?', default=None, const='raw')
		parser.opt('keep-edits', action='store_true')
		parser.opt('fzf', action='store_true')
		parser.opt('keep-db', action='store_true')
		parser.opt('reset-db', action='store_true')
		parser.opt('max-parallel-tasks', type=int, default=os.cpu_count())
		parser.opt('no-agg', action='store_true')
		parser.opt('no-breakpoint', action='store_true')
		parser.opt('no-mp', action='store_true')
		parser.opt('show-plots', action='store_true')
		parser.opt('show-plots-cmd', action='store_true')
		parser.opt('secrets', type=str)
		parser.opt('wait-pid', type=int, default=0)

		# Example
		parser.example('job job_file.yaml')
		parser.example('job job_part1.yaml job_part2.yaml')
		parser.example('job -keep-db -task last job_file.yaml')
		parser.example('job -show-plots job.yaml')
		parser.example('job -fzf -edit')
		return parser


	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		start_time = time()
		psys = platform.system().lower()
		logger = AmpelLogger.get_logger(base_flag=LogFlag.MANUAL_RUN)

		if args['no_mp']:
			"""
			# ThreadPool (dummy.Pool) is not enough in some circunstances
			# as matplotlib has issues when run outside main thread
			def Pool(**kwargs):
				kwargs['processes'] = 1
				return multiprocessing.dummy.Pool(**kwargs)
			"""
			import multiprocessing
			multiprocessing.__dict__['Pool'] = MockPool
			if not args['no_agg']:
				from ampel.util.getch import yes_no
				if yes_no('Set -no-agg option too (required for matplotlib interactions)'):
					args['no_agg'] = True

		if not args['no_agg']:
			try:
				import matplotlib as mpl
				mpl.use('Agg')
			except Exception:
				pass

		if args['no_breakpoint']:
			os.environ["PYTHONBREAKPOINT"] = "0"

		if isinstance(args['task'], (int, str)):
			args['task'] = [args['task']]

		schema_files = args['schema'] or [
			unknown_args.pop(unknown_args.index(el)) # type: ignore[attr-defined]
			for el in list(unknown_args)
			# This might conflict with very special - yet never used - config overrides
			if el.endswith('yml') or el.endswith('yaml')
		]

		if args.get('fzf') and not schema_files and psys.lower() != "windows":

			if shutil.which('fzf') is None:
				with out_stack():
					raise ValueError("Option '-fzf' requires the fzf command line utility")

			p1 = subprocess.Popen(
				['find', '.', '-name', '*.yml', '-o', '-name', '*.yaml'],
				stdout = subprocess.PIPE
			)
			p2 = subprocess.Popen('fzf', stdin=p1.stdout, stdout=subprocess.PIPE)
			out, err = p2.communicate()

			if err:
				with out_stack():
					raise ValueError("Error occured during fzf execution")

			if (selection := out.decode('utf8').strip()):
				schema_files = [selection]
			else:
				print("No schema selected")
				return

		if args.get('edit'):
			tmp_files: list[str] = []

		if args.get('edit') == 'raw':

			if schema_files:
				for sfile in list(schema_files):
					_, fname = tempfile.mkstemp(suffix='.yml', text=True)
					shutil.copyfile(sfile, fname)
					edit_job(fname)
					tmp_files.append(fname)

					if filecmp.cmp(sfile, fname):
						logger.info(f'Using original schema: {sfile}')
						os.unlink(fname)
						continue

					schema_files[schema_files.index(sfile)] = fname
					logger.info(f'Using modified schema: {fname}')
			else:
				schema_files = [
					edit_job(tempfile.mkstemp(suffix='.yml')[1])
				]

		if not schema_files:
			self.get_parser().print_help()
			return

		job, _ = self.get_job_schema(schema_files, logger, compute_sig=False)

		if args.get('edit') == 'parsed':
			fd, fname = tempfile.mkstemp(suffix='.yml')
			# Seems fd does not work with yaml.dump(), unsure why
			with open(fname, 'wt') as f:
				yaml.dump(
					ujson.loads(job.json(exclude_unset=True)), f,
					sort_keys=False, default_flow_style=None
				)
			edit_job(fname)
			tmp_files.append(fname)
			job, _ = self.get_job_schema([fname], logger, compute_sig=False)

		schema_descr = '|'.join(
			[os.path.basename(sf) for sf in schema_files]
		).replace('.yaml', '').replace('.yml', '')

		if len(schema_files) > 1:
			logger.info(f'Running job using composed schema: {schema_descr}')

		# Check or set env variable(s)
		for psys in ('any', platform.system().lower()):
			if psys in job.env:
				for k, v in job.env[psys].check.items():
					if k not in os.environ:
						logger.info(f'Environment variable {k}={v} required')
						return
					elif v and v != _maybe_int(os.environ[k]):
						logger.info(
							f'Environment variable {k} value mismatch:\n'
							f' required: {v} ({type(v)})\n'
							f' set: {os.environ[k]}'
						)
						return
				for k, v in job.env[psys].set.items():
					logger.info(f'Setting local environment variable {k}={v}')
					os.environ[k] = str(v)

		purge_db = job.mongo.reset or args['reset_db']
		if purge_db and args['keep_db']:
			logger.info('Keeping existing databases as option "-keep-db" was specified')
			purge_db = False

		if args['interactive']:

			from ampel.util.getch import yes_no

			try:
				if purge_db:
					purge_db = yes_no('Delete existing databases')

				args['task'] = []
				for i, td in enumerate(job.task):
					s = f' [{td.title}]' if td.title else ''
					if yes_no(f'Process task #{i}' + s):
						args['task'].append(i)
			except KeyboardInterrupt:
				sys.exit()

		if args['task']:

			if 'last' in args['task']:
				args['task'].remove('last')
				args['task'].append(len(job.task) - 1)

			if sum(args['task']) > 0 and not args['interactive'] and purge_db and not args['reset_db']:
				logger.info('Ampel job file requires db reset but argument "task" was provided')
				logger.info('Please add argument -reset-db to confirm you are absolutely sure...')
				return

		if job.requirements:
			dist_names = [el.lower() for el in get_dist_names()]
			missing = [dist for dist in job.requirements if dist.lower() not in dist_names]
			if missing:
				logger.info(f'Please install {try_reduce(missing)} to run this job\n')
				return # raise ValueError ?

		if not args['config'] and not os.path.exists(get_user_data_config_path()):
			from ampel.util.getch import yes_no
			if yes_no('Config seems to be missing, build and install'):
				from ampel.cli.ConfigCommand import ConfigCommand
				cc = ConfigCommand()
				cc.run({'install': True}, unknown_args=[], sub_op = 'build')

		s = f'Running job {job.name or schema_descr}'
		logger.info(s)

		print(' ' + '-'*len(s))

		# DevAmpelContext hashes automatically confid from potential IngestDirectives
		ctx = self.get_context(
			args, unknown_args, logger,
			freeze_config = False,
			ContextClass = DevAmpelContext,
			purge_db = purge_db,
			db_prefix = job.mongo.prefix,
			require_existing_db = False,
			one_db = True
		)

		config_dict = ctx.config._config
		self._patch_config(config_dict, job, logger)
		ctx.config._config = recursive_freeze(config_dict)

		# Ensure that job content saved in DB reflects options set dynamically
		if args['task']:
			for idx in sorted(args['task'], reverse=True):
				del job.task[idx]

		jtasks: list[dict[str, Any]] = []
		for i, model in enumerate(job.task):

			if isinstance(model, TemplateUnitModel):

				if model.template not in ctx.config._config['template']:
					raise ValueError(f'Unknown process template: {model.template}')

				fqn = ctx.config._config['template'][model.template]
				if ':' in fqn:
					fqn, class_name = fqn.split(':')
				else:
					class_name = fqn.split('.')[-1]
				Tpl = getattr(import_module(fqn), class_name)
				if not issubclass(Tpl, AbsProcessorTemplate):
					raise ValueError(f'Unexpected template type: {Tpl}')

				tpl = Tpl(**model.config)
				morphed_um = tpl \
					.get_model(ctx.config._config, model.dict()) \
					.dict(exclude_unset=True)

				if args.get('debug'):
					from ampel.util.pretty import prettyjson
					logger.info('Task model morphed by template:')
					for el in prettyjson(morphed_um, indent=4).split('\n'):
						logger.info(el)

				jtasks.append(morphed_um)

			else:
				jtasks.append(
					model.dict(
						exclude={'inputs', 'outputs', 'expand_with'},
						exclude_unset=True
					)
				)

			logger.info(
				f'Registering job task#{i} with ' +
				str(len(list(model.expand_with)) if model.expand_with else 1) +
				'x multiplier'
			)

		# recreate JobModel with templates resolved
		job_dict = ujson.loads(
			JobModel(
				**(
					job.dict(exclude_unset=True) | # type: ignore[arg-type]
					{
						'task': [
							td | task.dict(
								include={'inputs', 'outputs', 'expand_with', 'title'},
								exclude_unset=True
							)
							for task, td in zip(job.task, jtasks)
						]
					}
				)
			).json(exclude_unset=True)
		)

		if args.get('edit') == 'model':
			fd, fname = tempfile.mkstemp(suffix='.yml')
			# Seems fd does not work with yaml.dump(), unsure why
			with open(fname, 'wt') as f:
				yaml.dump(job_dict, f, sort_keys=False, default_flow_style=None)
			edit_job(fname)
			tmp_files.append(fname)
			with open(fname, 'rt') as f:
				job_dict = yaml.safe_load(f)
			jtasks = job_dict['task']

		if args.get('edit') and not args.get('keep_edits'):
			for el in tmp_files:
				try:
					os.unlink(el)
				except BaseException:
					pass

		if (wpid := args['wait_pid']) and psutil.pid_exists(wpid):
			logger.info(f'Waiting until process with PID {wpid} completes')
			while (psutil.pid_exists(wpid)):
				sleep(5)
			start_time = time()

		logger.info('Saving job schema')
		job_sig = build_unsafe_dict_id(job_dict, size=-64)
		ctx.db.get_collection('job').update_one(
			{'_id': job_sig},
			{'$setOnInsert': job_dict},
			upsert=True
		)

		run_ids = []
		for i, taskd in enumerate(jtasks):

			process_name = f'{job.name or schema_descr}#{i}'

			if 'title' in taskd:
				self.print_chapter(taskd['title'] if taskd.get('title') else f'Task #{i}', logger)
				#process_name += f' [{taskd['title']}]'
				del taskd['title']
			elif i != 0:
				self.print_chapter(f'Task #{i}', logger)

			if args['task'] is not None and i not in args['task']:
				logger.info(f'Skipping task #{i} as requested')
				continue

			taskd['override'] = taskd.pop('override', {}) | {'raise_exc': True}

			# Beacons have no real use in jobs (unlike prod)
			if taskd['unit'] == 'T2Worker' and 'send_beacon' not in taskd['config']:
				taskd['config']['send_beacon'] = False

			if (expand_with := job.task[i].expand_with) is not None:

				ps = []
				qs = []

				signal.signal(signal.SIGINT, signal_handler)
				signal.signal(signal.SIGTERM, signal_handler)

				try:
					for item in expand_with:

						self._fetch_inputs(job, job.task[i], item, logger)

						q: Queue = Queue()
						p = Process(
							target = run_mp_process,
							args = (
								q,
								config_dict,
								job.resolve_expressions(
									taskd,
									job.task[i],
									item
								),
								process_name,
							),
							daemon = True,
						)
						p.start()
						ps.append(p)
						qs.append(q)
					
					for i, (p, q) in enumerate(zip(ps, qs)):
						p.join()
						if (m := q.get()):
							logger.info(f'{taskd["unit"]}#{i} return value: {m}')
				except KeyboardInterrupt:
					sys.exit(1)
			
			else:
				
				self._fetch_inputs(job, job.task[i], None, logger)

				proc = ctx.loader.new_context_unit(
					model = UnitModel(**job.resolve_expressions(taskd, job.task[i])),
					context = ctx,
					process_name = process_name,
					job_sig = job_sig,
					sub_type = AbsEventUnit,
					base_log_flag = LogFlag.MANUAL_RUN
				)

				event_hdlr = EventHandler(
					proc.process_name,
					ctx.get_database(),
					raise_exc = proc.raise_exc,
					job_sig = job_sig,
					extra = {'task': i}
				)

				x = proc.run(event_hdlr)
				if event_hdlr.run_id:
					run_ids.append(event_hdlr.run_id)

				logger.info(f'{taskd["unit"]} return value: {x}')

		if len(run_ids) == 1:
			runstr = f" (run id: {run_ids[0]})"
		elif len(run_ids) > 1:
			runstr = " (run ids: " + " ".join([str(el) for el in run_ids]) + ")"
		else:
			runstr = ""

		logger.info(f'Job processed{runstr}')
		logger.info(f'Time required: {get_time_delta(start_time)}\n')

		if args.get('show_plots') or args.get('show_plots_cmd'):

			cmd = [
				'ampel', 'plot', 'show', '-stack', '300',
				'-png', os.environ.get('AMPEL_PLOT_DPI', '150'),
				'-t2', '-t3', '-base-path', 'body.plot', # to be improved later
				'-db', job.mongo.prefix, '-run-id', *[str(el) for el in run_ids],
			]

			if args.get('debug'):
				cmd.append('-debug')

			print(
				'%s\n%s command: %s' %
				('-' * 40, 'Executing' if args.get('show_plots') else 'Plot', " ".join(cmd))
			)

			if args.get('show_plots'):
				r = subprocess.run(cmd, check=True, capture_output=True, text=True)
				print(r.stdout)
				print(r.stderr)


	@staticmethod
	def print_chapter(msg: str, logger: AmpelLogger) -> None:
		logger.info(' ')
		logger.info('=' * (space := (len(msg) + 4)))
		logger.info('‖ ' + msg + ' ‖') # type: ignore
		logger.info('=' * space)
		logger.info(' ')


	@classmethod
	def get_job_schema(cls,
		schema_files: Sequence[str],
		logger: AmpelLogger,
		compute_sig: bool = True
	) -> tuple[JobModel, int]:

		lines = io.StringIO()
		for i, job_fname in enumerate(schema_files):

			if not os.path.exists(job_fname):
				with out_stack():
					raise FileNotFoundError(f'Job file not found: "{job_fname}"\n')

			with open(job_fname, 'r') as f:
				lines.write('\n'.join(f.readlines()))

		lines.seek(0)
		job = yaml.safe_load(lines)

		for k in list(job.keys()):
			# job root keys starting with % are used by own convention for yaml anchors
			# and thus need not be included in the loaded job structure
			if k.startswith('%'):
				del job[k]

		return JobModel(**job), build_unsafe_dict_id(job, size=-64) if compute_sig else 0


	@staticmethod
	def _fetch_inputs(
		job: JobModel,
		task: TaskUnitModel | TemplateUnitModel,
		item: None | str | dict | list,
		logger: AmpelLogger,
	):
		"""
		Ensure that input artifacts exist
		"""
		for artifact in task.inputs.artifacts:

			resolved_artifact = InputArtifact(
				**job.resolve_expressions(
					ujson.loads(artifact.json()), task, item
				)
			)

			if resolved_artifact.path.exists():
				logger.info(f'Artifact {resolved_artifact.name} exists at {resolved_artifact.path}')
			else:
				logger.info(
					f'Fetching artifact {resolved_artifact.name} from '
					f'{resolved_artifact.http.url} to {resolved_artifact.path}'
				)
				os.makedirs(resolved_artifact.path.parent, exist_ok=True)
				with tempfile.NamedTemporaryFile(delete=False) as tf:
					urlretrieve(resolved_artifact.http.url, tf.name)
					try:
						with tarfile.open(tf.name) as archive:
							logger.info(f'{resolved_artifact.name} is a tarball; extracting')
							os.makedirs(resolved_artifact.path)
							archive.extractall(resolved_artifact.path)
						os.unlink(tf.name)
					except tarfile.ReadError:
						os.rename(tf.name, resolved_artifact.path)


	@staticmethod
	def _patch_config(config_dict: dict[str, Any], job: JobModel, logger: AmpelLogger):

		# Add channel(s)
		for chan in job.channel:
			logger.info(f'Registering job channel "{chan.channel}"')
			dict.__setitem__(config_dict['channel'], chan.channel, chan.dict())

		# Add aliase(s)
		for k, v in job.alias.items():
			if 'alias' not in config_dict:
				dict.__setitem__(config_dict, 'alias', {})
			for kk, vv in v.items():
				logger.info(f'Registering job alias "{kk}"')
				if k not in config_dict['alias']:
					dict.__setitem__(config_dict['alias'], k, {})
				dict.__setitem__(config_dict['alias'][k], kk, vv)


def run_mp_process(
	queue: Queue,
	config: dict[str, Any],
	tast_unit_model: dict[str, Any],
	process_name: str,
	job_sig: None | int = None,
	task_nbr: None | int = None,
	log_profile: str = 'default'
) -> None:

	try:

		# Create new context with serialized config
		context = DevAmpelContext.load(config, one_db=True)

		processor = context.loader.new_context_unit(
			model = UnitModel(**tast_unit_model),
			context = context,
			sub_type = AbsEventUnit,
			log_profile = log_profile,
			process_name = process_name,
			job_sig = job_sig
		)

		queue.put(
			processor.run(
				EventHandler(
					processor.process_name,
					context.get_database(),
					raise_exc = processor.raise_exc,
					job_sig = job_sig,
					extra = {'task': task_nbr}
				)
			)
		)

	except Exception as e:
		import traceback
		se = str(e)
		queue.put(
			'\n' + '#'*len(se) + '\n' + str(e) + '\n' + '#'*len(se) + '\n' +
			''.join(traceback.format_exception(type(e), e, e.__traceback__))
		)


def signal_handler(sig, frame):
	#import traceback
	print('Interrupt detected')
	#print('Stack frames:')
	#traceback.print_stack(frame)
	raise KeyboardInterrupt()


def edit_job(fname: str) -> str:
	"""
	:raises ValueError if data edition fails
	"""
	if subprocess.call(
		os.environ.get('EDITOR', 'vi') + ' ' + fname,
		shell=True
	):
		with out_stack():
			raise ValueError("Cancellation requested during schema edition")

	return fname



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/T3BufferExporterStager.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.03.2021
# Last Modified Date:  17.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from io import BufferedWriter, TextIOWrapper
from datetime import datetime
from typing import BinaryIO, TextIO, Any
from collections.abc import Generator

from ampel.types import Traceless
from ampel.view.T3Store import T3Store
from ampel.log.AmpelLogger import AmpelLogger
from ampel.abstract.AbsT3Stager import AbsT3Stager
from ampel.abstract.AbsIdMapper import AbsIdMapper
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.content.T3Document import T3Document
from ampel.core.EventHandler import EventHandler
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.cli.export import txt_export, bin_export, get_fd

dsi = dict.__setitem__
ufts = datetime.utcfromtimestamp


class TextExportOptions(AmpelBaseModel):
	chunk_size: int = 200
	human_times: bool = True
	pretty: bool = False
	getch: bool = False


class T3BufferExporterStager(AbsT3Stager):
	"""
	Exports AmpelBuffer instances as BSON or text file
	"""

	logger: Traceless[AmpelLogger]
	event_hdlr: Traceless[EventHandler]
	channel: Any

	fd: None | str | BufferedWriter | TextIO | TextIOWrapper | BinaryIO
	binary: bool
	verbose: bool = True
	close_fd: bool = True
	id_mapper: None | AbsIdMapper = None
	txt_options: TextExportOptions = TextExportOptions()


	def stage(self,
		gen: Generator[AmpelBuffer, None, None],
		t3s: T3Store
	) -> None | Generator[T3Document, None, None]:

		fd, close_fd = get_fd(self.binary, self.fd, self.close_fd)
		
		if self.binary:
			bin_export(
				fd, # type: ignore[arg-type]
				gen, self.id_mapper, close_fd,
				AmpelLogger.get_logger() if self.verbose else None
			)

		else:
			txt_export(
				fd, # type: ignore[arg-type]
				gen, self.id_mapper, **self.txt_options.dict(), close_fd = self.close_fd,
				logger = AmpelLogger.get_logger() if self.verbose else None
			)

		return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/log/utils.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                30.09.2018
# Last Modified Date:  11.11.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys, traceback
from math import log2
from bson import ObjectId
from datetime import datetime
from typing import overload

from ampel.types import JDict
from ampel.core.AmpelDB import AmpelDB
from ampel.util.collections import has_nested_type
from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.LogFlag import LogFlag
from ampel.metrics.AmpelMetricsRegistry import AmpelMetricsRegistry
from ampel.protocol.LoggerProtocol import LoggerProtocol

exception_counter = AmpelMetricsRegistry.counter(
	"exceptions",
	"Number of exceptions caught and logged",
)

def log_exception(
	logger: LoggerProtocol,
	exc: None | Exception = None,
	extra: None | dict = None,
	last: bool = False,
	msg: None | str = None
) -> None:
	"""
	:param last: whether to print only the last exception in the stack
	:param msg: Optional message
	"""

	sys_exc = False

	if not exc:

		exc = getattr(sys, "last_value", None)
		sys_exc = True
		logger.error("loading exception from sys", extra=extra)

		if not exc:
			logger.error(
				"log_exception(..) was called but not exception could be found",
				extra=extra
			)
			return

	# Increment exception counter. Labels for process name and tier are added
	# implicitly in multiprocess mode.
	exception_counter.inc()

	logger.error("-" * 50, extra=extra)

	if msg:
		logger.error(msg)

	if last:
		exc.__context__ = None

	for el in traceback.format_exception(type(exc), exc, exc.__traceback__):
		for ell in el.split('\n'):
			if len(ell) > 0:
				logger.error(ell, extra=extra)

	logger.error("-" * 50, extra=extra)

	# Clear up recorded exception (avoiding potential multiple reports)
	if sys_exc:
		sys.last_value = None
		sys.last_traceback = None
		sys.last_type = None


def report_exception(
	ampel_db: AmpelDB,
	logger: AmpelLogger,
	exc: None | Exception = None,
	process: None | str = None,
	info: None | JDict = None
) -> None:
	"""
	:param tier: Ampel tier level (0, 1, 2, 3)
	:param logger: logger instance (logging module). \
	propagate_log() will be used to print details about the exception
	:param info: optional dict instance whose values will be included \
	in the document inserted into Ampel_troubles
	"""

	from traceback import format_exc
	from sys import exc_info

	# Don't create report for executions canceled manually
	if exc_info()[0] == KeyboardInterrupt:
		return

	# Feedback
	log_exception(logger, exc, info)

	trouble: JDict = {
		'_id': ObjectId(),
		'datetime': datetime.now().strftime('%d/%m/%Y %H:%M:%S'),
		'tier': get_tier_from_logger(logger)
	}

	# Logger with db_logging_handler must have a run id
	if db_logging_handler := logger.get_db_logging_handler():
		trouble['run'] = db_logging_handler.run_id

	# Additional info might have been provided (such as alert information)
	if info:
		trouble.update(info)

	if process:
		trouble['process'] = process

	trouble['exception'] = format_exc().replace("\"", "'").split("\n")

	# Populate 'troubles' collection
	insert_trouble(trouble, ampel_db, logger)


def report_error(
	ampel_db: AmpelDB,
	logger: AmpelLogger,
	msg: None | str = None,
	info: None | JDict = None
) -> None:
	"""
	This method is used to report bad states or errors which are grave enough
	to be worth the creation of a 'trouble document'.
	Information concerning the error can be provided as strine message through the 'msg' argument
	as well as dict through the parameter 'info'.
	This method should not be used to report Exceptions (please use report_exception(...))
	:raises: Should not raise errors
	"""

	# Increment exception counter
	exception_counter.inc()

	# Get filename and line number using inspect
	import inspect
	frame, filename, line_number, function_name, lines, index = inspect.stack()[1]

	trouble: dict[str, None | int | str | ObjectId] = {
		'_id': ObjectId(),
		'datetime': datetime.now().strftime('%d/%m/%Y %H:%M:%S'),
		'tier': get_tier_from_logger(logger),
		'location': '%s:%s' % (filename, line_number),
	}

	if msg:
		trouble['msg'] = msg

	# Additional info might have been provided (such as alert information)
	if info is not None:
		trouble.update(info)

	# Logger with db_logging_handler must have a run id
	if db_logging_handler := logger.get_db_logging_handler():
		trouble['run'] = db_logging_handler.run_id

	# Feedback
	logger.error("Error occured", extra=trouble)

	# Populate 'troubles' collection
	insert_trouble(trouble, ampel_db, logger)


def get_tier_from_logger(logger: AmpelLogger) -> None | int:

	lb = LogFlag(logger.base_flag)
	if LogFlag.T0 in lb:
		return 0
	elif LogFlag.T1 in lb:
		return 1
	elif LogFlag.T2 in lb:
		return 2
	elif LogFlag.T3 in lb:
		return 3

	return None


def get_tier_from_log_flags(flags: int | LogFlag) -> int:
	for i in (1, 2, 4, 8):
		if i & flags.__int__():
			return int(log2(i))
	return -1


def insert_trouble(
	trouble: JDict, ampel_db: AmpelDB, logger: AmpelLogger
) -> None:

	# Populate troubles collection
	try:
		ampel_db \
			.get_collection('trouble') \
			.insert_one(trouble)

	except Exception as e:

		# Bad luck (possible cause: DB offline)
		logger.error(
			msg = "Exception occured while populating 'troubles' collection",
			exc_info=e
		)

		logger.error(
			msg = f"Unpublished 'troubles' document: {str(trouble)}",
			exc_info=e
		)


def safe_query_dict(
	match: JDict,
	update: None | JDict = None,
	dict_key: None | str = 'query'
) -> JDict:
	u"""
	| Builds a dict that can be passed as "extra" parameter to instances of AmpelLogger.
	| Returned dict has the following structure:

	.. sourcecode:: python\n
		{
			"query": {
				"match": dict,
				"update": optional_dict
			}
		}

	Possibly embedded dollar signs in dict keys of parameters \
	"match" and "update" are replaced with the the unicode character \
	'Fullwidth Dollar Sign': ＄ (see docstring of :func:`convert_dollars \
	<ampel.log.LogUtils.convert_dollars>`)
	"""

	extra = {'match': convert_dollars(match)}

	if update:
		extra = {
			'match': convert_dollars(match),
			'update': convert_dollars(update)
		}
	else:
		extra = convert_dollars(match)

	return {dict_key: extra} if dict_key else extra


@overload
def convert_dollars(arg: JDict) -> JDict:
	...
@overload
def convert_dollars(arg: list[JDict]) -> list[JDict]:
	...
def convert_dollars(arg: JDict | list[JDict]) -> JDict | list[JDict]:
	"""
	MongoDB does not allow documents containing dollars in 'top level key' \
	(raises InvalidDocument). In order to log DB queries commands, we substitute \
	the dollar sign with the unicode character 'Fullwidth Dollar Sign': ＄.
	Another option would be do cast the dict to string (what we did before v0.5) \
	but it is less readable and takes more storage space.
	Nested dict shallow copies are performed.
	"""

	if isinstance(arg, dict):

		pblm_keys = [key for key in arg.keys() if "$" in key or "." in key]
		if pblm_keys:
			arg = arg.copy() # shallow copy
			for key in pblm_keys:
				if "$" in key:
					v = arg.pop(key)
					arg[(key := key.replace("$", "\uFF04"))] = v
				if "." in key:
					arg[key.replace(".", "\u2219")] = arg.pop(key)

		if not has_nested_type(arg, dict): # type: ignore[arg-type]
			return arg

		if not pblm_keys:
			arg = arg.copy()

		for key in arg.keys():
			arg[key] = convert_dollars(arg[key])

	elif isinstance(arg, list):
		if has_nested_type(arg, dict):
			arg = arg.copy() # type: ignore[attr-defined]
			return [convert_dollars(el) for el in arg] # type: ignore[arg-type]

	return arg



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/cli/BufferCommand.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.03.2021
# Last Modified Date:  27.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import sys
from argparse import ArgumentParser
from typing import Any
from collections.abc import Sequence
from ampel.log.LogFlag import LogFlag
from ampel.core.AmpelContext import AmpelContext
from ampel.cli.utils import maybe_load_idmapper
from ampel.cli.ArgParserBuilder import ArgParserBuilder
from ampel.cli.AbsStockCommand import AbsStockCommand
from ampel.cli.AbsLoadCommand import AbsLoadCommand
from ampel.cli.AmpelArgumentParser import AmpelArgumentParser
from ampel.t3.T3Processor import T3Processor

hlp = {
	'config': 'Path to an ampel config file (yaml/json)',
	# Optional
	'secrets': 'Path to a YAML secrets store in sops format',
	'db': 'Database prefix. If set, "-mongo.prefix" value will be ignored',
	'out': 'Path to file where serialized views will be saved (printed to stdout otherwise)',
	'log-profile': 'One of: default, compact, headerless, verbose, debug',
	'id-mapper': 'Convert stock ids using the provided id mapper (ex: ZTFIdMapper)',
	'binary': 'Store buffers as compact BSON structures',
	'pretty': 'Show buffers using spacier prettyjson',
	'getch': 'Wait for a key to be pressed in terminal before printing next buffer (not compatible with binary option)'
}

class BufferCommand(AbsStockCommand, AbsLoadCommand):


	@staticmethod
	def get_sub_ops() -> list[str]:
		return ['show', 'save']


	# Mandatory implementation
	def get_parser(self, sub_op: None | str = None) -> ArgumentParser | AmpelArgumentParser:

		if sub_op in self.parsers:
			return self.parsers[sub_op]

		sub_ops = self.get_sub_ops()
		if sub_op is None or sub_op not in sub_ops:
			return AmpelArgumentParser.build_choice_help(
				'buffer', sub_ops, hlp, description = 'Show or save ampel buffers.'
			)

		builder = ArgParserBuilder('buffer')
		hlp.update(self.get_select_args_help())
		hlp.update(self.get_load_args_help())
		builder.add_parsers(sub_ops, hlp)

		builder.req('config')
		builder.req('out', 'save', default=True)

		builder.opt('secrets', default=None)
		builder.opt('id-mapper')
		builder.opt('debug', action='store_true')
		builder.opt('log-profile', default='default')
		builder.opt('binary', 'save', action='store_true')
		builder.opt('pretty', 'show', action='store_true')
		builder.opt('getch', 'show', action='store_true')

		# Selection args
		self.add_selection_args(builder)

		# Content args
		self.add_load_args(builder, 'Views content arguments')

		# Example
		builder.example('show', '-stock 85628462 -no-t0')
		builder.example('show', '-db MyDB -stock 519059889 -no-t0')
		builder.example('save', '-stock 85628462 -no-t0 -out ZTFabcdef.json')
		
		self.parsers.update(
			builder.get()
		)

		return self.parsers[sub_op]


	# Mandatory implementation
	def run(self, args: dict[str, Any], unknown_args: Sequence[str], sub_op: None | str = None) -> None:

		ctx = self.get_context(
			args, unknown_args, ContextClass = AmpelContext,
			require_existing_db = args['db'] or True, one_db='auto'
		)

		maybe_load_idmapper(args)

		conf = {
			'fd': open(args['out'], 'wb' if args.get('binary') else 'w') \
				if sub_op == 'save' else sys.stdout,
			'id_mapper': args['id_mapper'],
			'verbose': sub_op == 'save',
			'binary': args.get('binary') or False
		}

		if args.get('pretty'):
			conf['pretty'] = True

		if not args.get('binary') and args.get('getch'):
			conf['getch'] = True

		t3p = T3Processor(
			context = ctx,
			process_name = 'BufferCommand',
			template = 'compact_t3',
			base_log_flag = LogFlag.MANUAL_RUN,
			log_profile = 'console_debug' if args.get('debug') else 'console_info',
			execute = [
				{
					'supply': {
						'unit': 'T3DefaultBufferSupplier',
						'config': {
							'select': self.build_select_model(args),
							'load': self.build_load_model(args)
						}
					},
					'stage': {
						'unit': 'T3BufferExporterStager',
						'config': conf
					}
				}
			]
		)

		t3p.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/demo/DemoReviewT3Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.06.2020
# Last Modified Date:  17.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator
from ampel.types import UBson, T3Send
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.struct.JournalAttributes import JournalAttributes
from ampel.struct.UnitResult import UnitResult
from ampel.view.SnapView import SnapView
from ampel.view.T3Store import T3Store


class DemoReviewT3Unit(AbsT3ReviewUnit[SnapView]):

	parameter: int = 10

	def process(self,
		gen: Generator[SnapView, T3Send, None],
		t3s: T3Store
	) -> UBson | UnitResult:

		self.logger.info(f"DemoReviewT3Unit output (parameter={self.parameter}):")

		for i, v in enumerate(gen, 1):

			self.logger.info("id: " + str(v.id))
			gen.send(
				# This journal customization will be applied only to the current 'stock'
				JournalAttributes(tag="DemoT3SpecificTag", extra={'a': 1})
			)

		# This journal customization will be applied to all stocks
		return UnitResult(
			body = {'param': 'value'},
			code = 10,
			journal = JournalAttributes(tag="DemoReviewT3UnitTag")
		)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/distrib.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                13.03.2021
# Last Modified Date:  23.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import os, re
from collections.abc import Generator
from pkg_resources import ( # type: ignore[attr-defined]
	get_distribution, AvailableDistributions,
	EggInfoDistribution, DistInfoDistribution
)


def get_dist_names(distrib_prefixes: list[str] = ["ampel-", "pyampel-"]) -> list[str]:
	""" Get all installed distributions whose names start with the provided prefix """
	# ensure that at least interface and core are found
	prefixes = {"ampel-interface", "ampel-core"}.union(distrib_prefixes)
	ret = [
		dist_name for dist_name in AvailableDistributions()
		if any(prefix in dist_name for prefix in prefixes)
	]

	if ret:
		ret.insert(0, ret.pop([i for i, el in enumerate(ret) if "interface" in el][0]))
		ret.insert(0, ret.pop([i for i, el in enumerate(ret) if "core" in el][0]))

	return ret


def get_files(
	dist_name: str,
	lookup_dir: None | str = None,
	pattern: None | re.Pattern = None
) -> list[str]:
	""" Loads all known conf files of the provided distribution (name) """

	if lookup_dir and not lookup_dir.endswith("/"):
		lookup_dir += "/"

	distrib = get_distribution(dist_name)

	# DistInfoDistribution: look for metadata RECORD (in <dist_name>.dist-info)
	if isinstance(distrib, DistInfoDistribution):

		# "pip3 install" in editable mode with pyproject.toml present
		# 1) grab the pth file
		if pth := next(
			(
				pth for el in distrib.get_metadata_lines('RECORD')
				if (pth := el.split(",")[0]).endswith(".pth")
			),
			None
		):

			fname = pth if os.path.isfile(pth) else distrib.get_resource_filename(__name__, pth)
			# 2) Manually look for files in referenced folder
			with open(fname, "r") as f:
				return list(
					walk_dir(
						f.read().strip(),
						lookup_dir,
						pattern
					)
				)

		# <pip3 install .> in non-editable mode with pyproject.toml present
		# Example of the ouput of distrib.get_metadata_lines('RECORD'):
		# 'conf/ampel-ztf.conf,sha256=FZkChNKKpcMPTO4pwyKq4WS8FAbznuR7oL9rtNYS7U0,322',
		# 'ampel/model/ZTFLegacyChannelTemplate.py,sha256=zVtv4Iry3FloofSazIFc4h8l6hhV-wpIFbX3fOW2njA,2182',
		# 'ampel/model/__pycache__/ZTFLegacyChannelTemplate.cpython-38.pyc,,',
		else:
			return [
				fname for el in distrib.get_metadata_lines('RECORD')
				if _check_match((fname := el.split(",")[0]), lookup_dir, pattern)
			]

	# "pip3 install" in editable mode without pyproject.toml present
	elif isinstance(distrib, EggInfoDistribution):
		# Example of the ouput of distrib.get_metadata_lines('SOURCES.txt'):
		# 'setup.py',
		# 'conf/ampel-ztf.json',
		# 'ampel/model/ZTFLegacyChannelTemplate.py',
		return [
			el for el in distrib.get_metadata_lines('SOURCES.txt')
			if _check_match(el, lookup_dir, pattern)
		]

	else:
		raise ValueError(f"Unsupported distribution type: '{type(distrib)}'")


def _check_match(
	arg: str,
	lookup_dir: None | str,
	pattern: None | re.Pattern
) -> bool:
	if lookup_dir and not arg.startswith(lookup_dir):
		return False
	if pattern and not pattern.match(arg):
		return False
	return True


def walk_dir(
	path: str,
	lookup_dir: None | str = None,
	pattern: None | re.Pattern = None
) -> Generator[str, None, None]:
	for root, dirs, files in os.walk(f"{path}/{lookup_dir}" if lookup_dir else f"{path}"):
		for fname in files:
			if pattern and not pattern.match(fname):
				continue
			yield os.path.join(root, fname)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/T3DocBuilder.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                08.12.2021
# Last Modified Date:  28.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from datetime import datetime
from typing import Any, Union
from importlib import import_module
from collections.abc import Iterable

from ampel.types import Traceless, StockId, UBson, ubson
from ampel.abstract.AbsT3ReviewUnit import AbsT3ReviewUnit
from ampel.abstract.AbsT3PlainUnit import AbsT3PlainUnit
from ampel.abstract.AbsT3ControlUnit import AbsT3ControlUnit
from ampel.abstract.AbsUnitResultAdapter import AbsUnitResultAdapter
from ampel.model.t3.T3DocBuilderModel import T3DocBuilderModel
from ampel.log.AmpelLogger import AmpelLogger
from ampel.core.ContextUnit import ContextUnit
from ampel.core.EventHandler import EventHandler
from ampel.content.T3Document import T3Document
from ampel.content.MetaRecord import MetaRecord
from ampel.struct.UnitResult import UnitResult
from ampel.enum.DocumentCode import DocumentCode
from ampel.enum.MetaActionCode import MetaActionCode
from ampel.enum.JournalActionCode import JournalActionCode
from ampel.mongo.update.MongoStockUpdater import MongoStockUpdater
from ampel.view.T3Store import T3Store
from ampel.util.mappings import dictify
from ampel.util.tag import merge_tags
from ampel.util.hash import build_unsafe_dict_id

AbsT3s = Union[AbsT3ControlUnit, AbsT3ReviewUnit, AbsT3PlainUnit]


class T3DocBuilder(ContextUnit, T3DocBuilderModel):
	"""
	Provides methods for handling UnitResult and generating a T3Document out of it
	"""

	logger: Traceless[AmpelLogger]
	event_hdlr: Traceless[EventHandler]


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)
		self.adapters: dict[str, AbsUnitResultAdapter] = {}
		self.stock_updr = MongoStockUpdater(
			ampel_db = self.context.db,
			tier = 3,
			run_id = self.event_hdlr.get_run_id(),
			process_name = self.event_hdlr.process_name,
			logger = self.logger,
			raise_exc = self.event_hdlr.raise_exc,
			extra_tag = self.extra_journal_tag,
			update_journal = self.update_journal,
			bump_updated = False
		)


	def handle_t3_result(self,
		t3_unit: AbsT3s,
		res: UBson | UnitResult,
		t3s: T3Store,
		stocks: None | list[StockId],
		ts: float
	) -> None | T3Document:

		# Let's consider logs as a result product
		if (buf_hdlr := getattr(t3_unit, '_buf_hdlr')) and buf_hdlr.buffer:
			buf_hdlr.forward(self.logger)

		if isinstance(res, UnitResult):
			if stocks and res.journal:
				self.stock_updr.add_journal_record(
					stock = stocks, # used to match stock docs
					jattrs = res.journal,
					unit = t3_unit.__class__.__name__,
					action_code = JournalActionCode.T3_ADD_DOC
				)
			if res.body is not None or res.code is not None:
				return self.craft_t3_doc(t3_unit, res, t3s, ts, stocks)
		elif res is not None or (res is None and self.save_stock_ids and stocks):
			return self.craft_t3_doc(t3_unit, res, t3s, ts, stocks)

		return None


	def craft_t3_doc(self,
		t3_unit: AbsT3s,
		res: None | UBson | UnitResult,
		t3s: T3Store,
		ts: float,
		stocks: None | list[StockId] = None
	) -> T3Document:

		t3d: T3Document = {'process': self.event_hdlr.process_name} # type: ignore[typeddict-item]
		actact = MetaActionCode(0)
		now = datetime.now()

		if self.human_timestamp:
			t3d['datetime'] = now.strftime(self.human_timestamp_format)

		t3d['unit'] = t3_unit.__class__.__name__
		t3d['code'] = actact

		conf = t3_unit._get_trace_content()
		meta: MetaRecord = {
			'run': self.event_hdlr.get_run_id(),
			'ts': int(now.timestamp()),
			'duration': time() - ts
		}

		confid = build_unsafe_dict_id(conf)
		self.context.db.add_conf_id(confid, conf)

		# Live dangerously
		if confid not in self.context.config._config['confid']:
			dict.__setitem__(self.context.config._config['confid'], confid, conf)

		t3d['confid'] = confid

		if self.resolve_config:
			t3d['config'] = conf

		if self.channel:
			t3d['channel'] = self.channel
			actact |= MetaActionCode.ADD_CHANNEL

		if self.save_stock_ids and stocks:
			t3d['stock'] = stocks

		t3d['code'] = DocumentCode.OK
		t3d['meta'] = meta # note: mongodb maintains key order

		if t3s.session:
			t3d['session'] = dictify(t3s.session)

		if isinstance(res, UnitResult):

			if res.code:
				t3d['code'] = res.code
				actact |= MetaActionCode.SET_UNIT_CODE
			else:
				actact |= MetaActionCode.SET_CODE

			if res.tag:
				if self.tag:
					t3d['tag'] = merge_tags(self.tag, res.tag) # type: ignore
				else:
					t3d['tag'] = res.tag
			elif self.tag:
				t3d['tag'] = self.tag

			if res.body:

				if res.adapter:
					if res.adapter not in self.adapters:
						self.adapters[res.adapter] = getattr(
							import_module(f"ampel.core.adapter.{res.adapter}"),
							res.adapter
						)(context=self.context, run_id=self.event_hdlr.get_run_id())
					res = self.adapters[res.adapter].handle(res)

				t3d['body'] = res.body
				actact |= MetaActionCode.ADD_BODY

		else:

			if self.tag:
				t3d['tag'] = self.tag

			# bson
			if isinstance(res, ubson):
				t3d['body'] = res
				actact |= (MetaActionCode.ADD_BODY | MetaActionCode.SET_CODE)

			else:
				actact |= MetaActionCode.SET_CODE

		meta['activity'] = [{'action': actact}]

		if self.human_id:
			ids = []
			if 'process' in self.human_id:
				ids.append("[%s]" % self.event_hdlr.process_name)
			if 'taskindex' in self.human_id:
				ids.append("[#%s]" % self.event_hdlr.process_name.split("#")[-1])
			if 'unit' in self.human_id:
				ids.append("[%s]" % t3_unit.__class__.__name__)
			if 'tag' in self.human_id and 'tag' in t3d:
				ids.append("[%s]" % (t3d['tag'] if isinstance(t3d['tag'], (int, str)) else " ".join(t3d['tag']))) # type: ignore[arg-type]
			if 'config' in self.human_id:
				ids.append("[%s]" % build_unsafe_dict_id(conf))
			if 'run' in self.human_id:
				ids.append("[%s]" % self.stock_updr.run_id) # not great
			ids.append(now.strftime(self.human_timestamp_format))
			t3d['_id'] = " ".join(ids)

		return t3d


	def flush(self, arg: AbsT3s | Iterable[AbsT3s], extra: None | dict[str, Any] = None) -> None:

		for t3_unit in [arg] if isinstance(arg, (AbsT3ControlUnit, AbsT3ReviewUnit, AbsT3PlainUnit)) else arg:

			if (handlers := getattr(t3_unit.logger, 'handlers')) and handlers[0].buffer:
				handlers[0].forward(self.logger, extra=extra)
				self.logger.break_aggregation()

			if self.stock_updr.update_journal:
				self.stock_updr.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/collector/DBConfigCollector.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.10.2019
# Last Modified Date:  20.04.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from ampel.mongo.model.AmpelDBModel import AmpelDBModel
from ampel.config.collector.ConfigCollector import ConfigCollector
from ampel.config.collector.AbsDictConfigCollector import AbsDictConfigCollector
from ampel.log import VERBOSE


class DBConfigCollector(AbsDictConfigCollector):

	def add(self,
		arg: dict[str, Any],
		dist_name: str,
		version: str | float | int,
		register_file: str
	) -> None:

		# Allow 'mongo': {'prefix': ...} in general ampel.conf
		if len(arg) == 1 and 'prefix' in arg:
			self.__setitem__('prefix', arg['prefix'])
			return

		# Allow 'mongo': {'ingest': ...} in general ampel.conf
		if len(arg) == 1 and 'ingest' in arg:
			self.__setitem__('ingest', arg['ingest'])
			return

		# At this point, arg usually contains content of files
		# contained in pyampel-core/conf/ampel-core/mongo/*

		dbs = self.get('databases')

		if not dbs:
			dbs = []
			self.__setitem__('databases', dbs)
			if 'prefix' not in self:
				self.__setitem__('prefix', 'Ampel')


		# validate model
		if self.verbose:
			self.logger.log(VERBOSE, 'Validating DB configuration')

		try:

			m = AmpelDBModel(**arg)

			if self.verbose:
				self.logger.log(VERBOSE, f'Configuration of DB collection "{m.name}" is valid')

			if m.name not in [db['name'] for db in dbs]:
				dbs.append(arg)
			else:
				for db in dbs:
					if db['name'] == m.name:
						if db['role'] != m.role:
							raise ValueError(
								f"Incompatible database configurations for {m.name} (role mismatch)"
							)
						common_cols = {col['name'] for col in db['collections']} & {col.name for col in m.collections}
						if common_cols:
							raise ValueError(
								f"Incompatible database configurations for {m.name} (collection already defined)"
							)
						for col_def in arg['collections']:
							db['collections'].append(col_def)
						break

		except Exception as e:

			self.error(
				'Incorrect DB configuration ' +
				ConfigCollector.distrib_hint(dist_name, register_file),
				exc_info = e
			)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/db/AmpelDBModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                19.10.2019
# Last Modified Date:  08.03.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Sequence
from ampel.mongo.model.AmpelColModel import AmpelColModel
from ampel.mongo.model.MongoClientRoleModel import MongoClientRoleModel
from ampel.base.AmpelBaseModel import AmpelBaseModel

class AmpelDBModel(AmpelBaseModel):
	name: str
	collections: Sequence[AmpelColModel]
	role: MongoClientRoleModel



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/aux/filter/SimpleDictArrayFilter.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.01.2020
# Last Modified Date:  18.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import TypeVar
from collections.abc import Mapping, Sequence
from ampel.model.aux.FilterCriterion import FilterCriterion
from ampel.aux.filter.AbsLogicOperatorFilter import AbsLogicOperatorFilter

T = TypeVar("T", bound=Mapping)


class SimpleDictArrayFilter(AbsLogicOperatorFilter[T]):
	"""
	In []: f = SimpleDictArrayFilter(filters={'attribute': 'a', 'operator': '==', 'value': 2})
	In []: f.apply([{'a': 1}, {'a': 4}])
	Out[]: []

	In []: f.apply([{'a': 1}, {'a': 2}])
	Out[]: [{'a': 2}]

	In []: f1 = SimpleDictArrayFilter(filters={'attribute': 'a', 'type': 'Sequence', 'operator': 'contains', 'value': 2})
	In []: f1.apply([{'a': [1, 2]}, {'a': 4}])
	Out[]: [{'a': [1, 2]}]

	In []: f2 = SimpleDictArrayFilter(filters={
		'any_of': [
			{'attribute': 'a', 'type': 'Sequence', 'operator': 'contains', 'value': 2},
			{'attribute': 'a', 'type': 'Sequence', 'operator': 'contains', 'value': 1}
		]
	})
	In []: f2.apply([{'a': [1, 2], 'b': {'f': 12}}, {'a': 4}, {'a': 2, 'z': 12}])
	Out[]: [{'a': [1, 2], 'b': {'f': 12}}]
	"""

	@staticmethod
	def _apply_filter(dicts: Sequence[T], f: FilterCriterion) -> list[T]:

		attr_name = f.attribute
		if f.type:
			return [
				d for d in dicts
				if attr_name in d and
				isinstance(d[attr_name], f.type) and
				f.operator(d[attr_name], f.value)
			]
		else:
			return [d for d in dicts if attr_name in d and f.operator(d[attr_name], f.value)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/supply/T3DefaultBufferSupplier.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.07.2021
# Last Modified Date:  13.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from collections.abc import Generator, Sequence
from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.abstract.AbsT3Selector import AbsT3Selector
from ampel.abstract.AbsT3Loader import AbsT3Loader
from ampel.abstract.AbsBufferComplement import AbsBufferComplement
from ampel.util.collections import get_chunks as chunks_func
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.model.UnitModel import UnitModel
from ampel.view.T3Store import T3Store


class T3DefaultBufferSupplier(AbsT3Supplier[Generator[AmpelBuffer, None, None]]):
	"""
	Default T3 supplier that:
	- matches stock ids based on the stock collection (section 'select'),
	- loads AmpelBuffer content as defined in the 'load' section
	- potentially updates AmpelBuffer through the 'complement' section
	"""

	#: Select stocks
	select: UnitModel

	#: Fill :class:`~ampel.core.AmpelBuffer.AmpelBuffer` for each selected stock
	load: UnitModel

	#: Add external information to each :class:`~ampel.core.AmpelBuffer.AmpelBuffer`.
	complement: None | Sequence[UnitModel]

	#: number of stocks to load at once. Set to 0 to disable chunking
	chunk_size: int = 1000


	def __init__(self, **kwargs) -> None:

		super().__init__(**kwargs)

		# stock selection
		#################

		# Spawn and run a new selector instance
		# fetch() returns an iterable (often a pymongo cursor)
		self.selector = self.context.loader.new_context_unit(
			model = self.select,
			context = self.context,
			sub_type = AbsT3Selector,
			logger = self.logger
		)

		# Content loader
		################

		# Spawn requested content loader
		self.data_loader = self.context.loader \
			.new_context_unit(
				model = self.load,
				context = self.context,
				sub_type = AbsT3Loader,
				logger = self.logger
			)

		# Content complementer
		######################

		if self.complement:

			# Spawn requested snapdata complementers
			self.complementers: None | list[AbsBufferComplement] = [
				self.context.loader \
					.new_context_unit(
						model = conf_el,
						context = self.context,
						sub_type = AbsBufferComplement,
						logger = self.logger
					)
				for conf_el in self.complement
			]
		else:
			self.complementers = None


	def supply(self, t3s: T3Store) -> Generator[AmpelBuffer, None, None]:

		# NB: we consume the entire cursor at once using list() to be robust
		# against cursor timeouts or server restarts during long lived T3 processes
		stock_ids = list(self.selector.fetch() or [])
		if not stock_ids:
			return

		# Usually, id_key is '_id' but it can be 'stock' if the
		# selection is based on t2 documents for example
		id_key = self.selector.field_name

		# Run start
		###########
		chunks = chunks_func(stock_ids, self.chunk_size) if self.chunk_size > 0 else [stock_ids]

		# Loop over chunks from the cursor/iterator
		for chunk_ids in chunks:

			# allow working chunks to complete even if some raise exception
			try:

				# Load info from DB
				tran_data = self.data_loader.load([sid[id_key] for sid in chunk_ids])

				# Potentialy add complementary information (spectra, TNS names, ...)
				if self.complementers:
					for appender in self.complementers:
						appender.complement(tran_data, t3s)

				for ampel_buffer in tran_data:
					yield ampel_buffer

			except Exception as e:
				self.event_hdlr.handle_error(e, self.logger)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/TemplateUnitModel.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  13.08.2022
# Last Modified By:    jvs

from typing import Any
from pydantic import validator, root_validator
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.job.TaskInputs import TaskInputs
from ampel.model.job.TaskOutputs import TaskOutputs
from ampel.model.job.utils import ExpandWith, _parse_multiplier


class TemplateUnitModel(AmpelBaseModel):

    template: str
    title: str = ""
    config: dict[str, Any]
    inputs: TaskInputs = TaskInputs()
    outputs: TaskOutputs = TaskOutputs()
    expand_with: ExpandWith = None

    @validator("title", pre=True)
    def populate_title(cls, v, values):
        return v or values["template"]

    @root_validator(pre=True)
    def parse_multiplier(cls, values):
        return _parse_multiplier(values)



from typing import Any, Optional
from ampel.base.LogicalUnit import LogicalUnit
from ampel.model.ingest.CompilerOptions import CompilerOptions
from ampel.secret.NamedSecret import NamedSecret
from ampel.dev.DevAmpelContext import DevAmpelContext
import pytest

from ampel.secret.Secret import Secret
from ampel.secret.DictSecretProvider import DictSecretProvider
from ampel.secret.AmpelVault import AmpelVault
from ampel.model.UnitModel import UnitModel
from ampel.abstract.AbsOpsUnit import AbsOpsUnit
from ampel.core.ContextUnit import ContextUnit
from ampel.abstract.AbsEventUnit import AbsEventUnit


@pytest.fixture
def secrets():
    return DictSecretProvider(
        {
            "dict": {"a": 1, "b": "foo"},
            "str": "blahblablah",
            "tuple": (1, 1),
        }
    )


@pytest.mark.parametrize(
    "expected_type,key", [(dict, "dict"), (str, "str"), (tuple, "tuple")]
)
def test_resolve_secrets_correct_type(
    dev_context: DevAmpelContext,
    secrets: DictSecretProvider,
    monkeypatch,
    expected_type,
    key,
    ampel_logger,
):
    class Modelo(LogicalUnit):
        seekrit: NamedSecret[expected_type] = NamedSecret(label=key)  # type: ignore[valid-type]

    dev_context.register_unit(Modelo)
    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))

    unit = dev_context.loader.new(
        UnitModel(unit="Modelo"), logger=ampel_logger, unit_type=Modelo
    )
    assert unit.seekrit.get() == secrets.store[key]


@pytest.mark.parametrize(
    "expected_type,key", [(dict, "str"), (str, "tuple"), (tuple, "dict")]
)
def test_resolve_secrets_wrong_type(
    secrets, dev_context: DevAmpelContext, monkeypatch, expected_type, key, ampel_logger
):
    class Modelo(LogicalUnit):
        seekrit: NamedSecret[expected_type] = NamedSecret(label=key)  # type: ignore[valid-type]

    dev_context.register_unit(Modelo)
    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))
    with pytest.raises(ValueError):
        dev_context.loader.new(
            UnitModel(unit="Modelo"), logger=ampel_logger, unit_type=Modelo
        )


def test_resolve_secret_from_superclass(
    secrets: DictSecretProvider, dev_context: DevAmpelContext, monkeypatch, ampel_logger
):
    """Secrets are resolved with multiple inheritance"""

    class Base(ContextUnit):
        seekrit: NamedSecret[dict]

    class Derived(AbsOpsUnit, Base):
        def run(self, beacon):
            ...

    assert (
        "seekrit" not in Derived.__annotations__
    ), "multiply-inherited AmpelBaseModels are missing annotations"

    dev_context.register_unit(Derived)  # type: ignore[arg-type]
    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))
    unit = dev_context.loader.new(
        UnitModel(unit="Derived", config={"seekrit": {"label": "dict"}}),
        logger=ampel_logger,
        context=dev_context,
        unit_type=Derived,
    )
    assert unit.seekrit.get() == secrets.store.get("dict")


def test_resolve_secret_from_config(
    secrets, dev_context: DevAmpelContext, monkeypatch, ampel_logger
):
    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))

    class NiceAndConcrete(LogicalUnit):
        seekrit: NamedSecret[dict]

    dev_context.register_unit(NiceAndConcrete)
    # unit with concrete secret field can be instantiated
    dev_context.loader.new(
        UnitModel(unit="NiceAndConcrete", config={"seekrit": {"label": "dict"}}),
        logger=ampel_logger,
        unit_type=NiceAndConcrete,
    )

    # and also validated without instantiating
    with dev_context.loader.validate_unit_models():
        UnitModel(unit="NiceAndConcrete", config={"seekrit": {"label": "dict"}})
        with pytest.raises(TypeError):
            UnitModel(unit="NiceAndConcrete")

    # unit with abstract secret field cannot be instantiated
    class BadAndAbstract(LogicalUnit):
        seekrit: Secret[dict]

    dev_context.register_unit(BadAndAbstract)
    with pytest.raises(TypeError):
        dev_context.loader.new(
            UnitModel(unit="BadAndAbstract", config={"seekrit": {"label": "dict"}}),
            logger=ampel_logger,
            unit_type=BadAndAbstract,
        )


@pytest.mark.parametrize("config", [None, {"seekrit": {"label": "dict"}}])
def test_use_secret_in_init(
    secrets: DictSecretProvider,
    dev_context: DevAmpelContext,
    monkeypatch,
    ampel_logger,
    config,
):
    """Secrets are populated before being passed to init"""
    monkeypatch.setattr(dev_context.loader, "vault", AmpelVault(providers=[secrets]))

    class NeedsSecretInInit(LogicalUnit):
        seekrit: NamedSecret[dict] = NamedSecret(label="dict")

        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            assert (
                self.seekrit.get() == secrets.store[self.seekrit.label]
            ), "secret is populated"

    dev_context.register_unit(NeedsSecretInInit)
    # secret field is populated
    dev_context.loader.new(
        UnitModel(unit="NeedsSecretInInit", config=config),
        logger=ampel_logger,
        unit_type=NeedsSecretInInit,
    )


def test_unit_validation(dev_context: DevAmpelContext):
    class Dummy(LogicalUnit):
        param: int = 42

    dev_context.register_unit(Dummy)

    with dev_context.loader.validate_unit_models():
        # simple, one-level validation
        UnitModel(unit="Dummy")
        UnitModel(unit="Dummy", config={"param": 37})
        with pytest.raises(TypeError):
            UnitModel(unit="Dummy", config={"param": "fish"})
        with pytest.raises(TypeError):
            UnitModel(unit="Dummy", config={"nonexistant_param": True})

        t3_config: dict[str, Any] = {
            "execute": [
                {
                    "unit": "T3ReviewUnitExecutor",
                    "config": {
                        "supply": {
                            "unit": "T3DefaultBufferSupplier",
                            "config": {
                                "select": {"unit": "T3StockSelector"},
                                "load": {
                                    "unit": "T3SimpleDataLoader",
                                    "config": {"directives": [{"col": "stock"}]},
                                },
                            },
                        },
                        "stage": {
                            "unit": "T3SimpleStager",
                            "config": {"execute": [{"unit": "DemoReviewT3Unit"}]},
                        },
                    },
                }
            ],
        }

        # recursive validation
        UnitModel(unit="T3Processor", config=t3_config)

        with pytest.raises(TypeError):
            t3_config["execute"][0]["config"]["supply"]["config"]["select"][
                "unit"
            ] = "NotActuallyAUnit"
            UnitModel(unit="T3Processor", config=t3_config)


def test_compiler_options_validation(mock_context: DevAmpelContext):
    """AuxAliasableUnit can be intialized from a string"""

    class Dummy(LogicalUnit):
        compiler_options: Optional[CompilerOptions]

    mock_context.register_unit(Dummy)

    with mock_context.loader.validate_unit_models():
        UnitModel(unit="Dummy")
        UnitModel(unit="Dummy", config={"compiler_options": "DummyCompilerOptions"})
        with pytest.raises(TypeError):
            UnitModel(unit="Dummy", config={"compiler_options": "foo"})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/ProcessModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.10.2019
# Last Modified Date:  30.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

import schedule as sched
from typing import Literal
from collections.abc import Sequence
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.types import ChannelId
from ampel.model.UnitModel import UnitModel
from ampel.config.ScheduleEvaluator import ScheduleEvaluator


class ProcessModel(AmpelBaseModel):

	name: str
	version: int | float | str
	active: bool = True
	tier: None | Literal[0, 1, 2, 3]
	schedule: Sequence[str]
	channel: None | ChannelId | Sequence[ChannelId]
	distrib: None | str
	source: None | str
	isolate: bool = True
	multiplier: int = 1
	log: None | str
	controller: UnitModel = UnitModel(unit='DefaultProcessController')
	processor: UnitModel


	def __init__(self, **kwargs) -> None:

		if isinstance(kwargs.get('schedule'), str):
			kwargs['schedule'] = [kwargs['schedule']]

		super().__init__(**kwargs)

		evaluator = None
		for el in self.schedule:
			if el == "super":
				continue
			try:
				if evaluator is None:
					evaluator = ScheduleEvaluator()
				evaluator(sched.Scheduler(), el).do(lambda x: None)
			except Exception:
				raise ValueError("Incorrect 'schedule' parameter")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/job/OutputParameter.py
# License:             BSD-3-Clause
# Author:              jvs
# Date:                Unspecified
# Last Modified Date:  13.08.2022
# Last Modified By:    jvs

from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.job.OutputParameterSource import OutputParameterSource

class OutputParameter(AmpelBaseModel):

    name: str
    value_from: OutputParameterSource

    def value(self) -> None | str:
        try:
            return self.value_from.path.read_text()
        except FileNotFoundError:
            return self.value_from.default



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/secret/AEAbsSecretProvider.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                20.06.2021
# Last Modified Date:  20.06.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from sjcl import SJCL
from typing import Union
from collections.abc import Iterable
from ampel.abstract.AbsSecretProvider import AbsSecretProvider
from ampel.secret.AESecret import AESecret
from ampel.secret.Secret import Secret


class AESecretProvider(AbsSecretProvider):

	def __init__(self, pwds: str | Iterable[str]):
		self.sjcl = SJCL()
		self.pwds = [pwds] if isinstance(pwds, str) else pwds


	def tell(self, arg: Secret, ValueType: type) -> bool:
		"""
		Potentially update an initialized Secret instance with
		the actual sensitive information associable with it.
		:returns: True if the Secret was told/resolved or
		False if the provided Secret is unknown to this secret provider
		"""

		if not issubclass(str, ValueType):
			return False

		if isinstance(arg, AESecret):
			for pwd in self.pwds:
				try:
					arg.set(
						self.sjcl \
							.decrypt(arg.dict(), pwd) \
							.decode("utf-8")
					)
					return True
				except Exception:
					continue

		return False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/builder/RemoteUnitDefinition.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.11.2019
# Last Modified Date:  15.06.2020
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import List
from ampel.base.AmpelBaseModel import AmpelBaseModel

class RemoteUnitDefinition(AmpelBaseModel):

	class_name: str
	base: list['str']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/demo/DemoPointT2Unit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                25.03.2020
# Last Modified Date:  15.12.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from time import time
from random import randint
from ampel.types import UBson
from ampel.struct.UnitResult import UnitResult
from ampel.content.DataPoint import DataPoint
from ampel.abstract.AbsPointT2Unit import AbsPointT2Unit


class DemoPointT2Unit(AbsPointT2Unit):

	test_parameter: int = 1
	chatty: bool = False

	def process(self, datapoint: DataPoint) -> UBson | UnitResult:

		ret = {
			"datapoint_id": datapoint['id'],
			"time": time(),
			"test_parameter": self.test_parameter
		}

		if self.chatty:
			self.logger.info(f"Parsing datapoint {datapoint['id']} (test_parameter={self.test_parameter})")

		if randint(0, 1) > 0.5:
			return ret
		else:
			return UnitResult(code=10, tag="MY_CUSTOM_TAG", body=ret)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/core/ContextUnit.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                07.10.2019
# Last Modified Date:  09.01.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.types import Traceless
from ampel.base.AmpelUnit import AmpelUnit
from ampel.core.AmpelContext import AmpelContext


class ContextUnit(AmpelUnit):
	"""
	Base class for units requiring a reference to an AmpelContext instance
	"""

	context: Traceless[AmpelContext]

	#: Private variable potentially set by UnitLoader for provenance purposes. Either:
	#: * None if provanance flag is False
	#: * 0 in case model content is not serializable
	#: * any other signed int value
	_trace_id: None | int = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/dev/NoShaper.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                05.08.2021
# Last Modified Date:  05.08.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from collections.abc import Iterable
from ampel.types import StockId
from ampel.abstract.AbsT0Unit import AbsT0Unit
from ampel.content.DataPoint import DataPoint
from ampel.log.AmpelLogger import AmpelLogger


class NoShaper(AbsT0Unit):

	# override
	logger: None | AmpelLogger # type: ignore[assignment]

	# Mandatory implementation
	def process(self, arg: Iterable[dict[str, Any]], stock: StockId) -> list[DataPoint]: # type: ignore[override]
		return arg # type: ignore



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                ampel/logging/LoggingErrorReporter.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                29.09.2018
# Last Modified Date:  18.01.2019
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.log.AmpelLogger import AmpelLogger
from ampel.log.utils import log_exception, report_exception

class LoggingErrorReporter:

	@staticmethod
	def report(handler, e, bwe_details=None):
		"""
		:param DBLoggingHandler handler:
		:param Exception e:
		:param BulkWriteError bwe_details:
		"""

		# Print log stack using std logging
		logger = AmpelLogger.get_logger()

		log_exception(logger, e, msg="Primary exception:")

		if bwe_details:
			logger.error("BulkWriteError details:")
			logger.error(bwe_details)
			logger.error("#" * 52)

		logger.error("DB log flushing error, un-flushed (json) logs below.")
		logger.error("*" * 52)

		for d in handler.log_dicts:
			logger.error(str(d))
		logger.error("#" * 52)

		try:
			# This will fail as well if we have DB connectivity issues
			report_exception(
				logger, e, run_id=handler.get_run_id(),
				info = None if bwe_details is None else {'BulkWriteError': str(bwe_details)}
			)
		except Exception as ee:
			log_exception(
				logger, ee, last=True,
				msg="Could not update troubles collection as well (DB offline?)"
			)

		# TODO: try slack ? (will fail if network issue)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/config/builder/AbsConfigTemplate.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                14.04.2020
# Last Modified Date:  18.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from ampel.base.AmpelABC import AmpelABC
from ampel.base.AmpelBaseModel import AmpelBaseModel


class AbsConfigTemplate(AmpelABC, AmpelBaseModel, abstract=True):
	""" Known direct subclasses: AbsProcessTemplate, AbsChannelTemplate """

	template: None | str



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/model/ChannelModel.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                09.10.2019
# Last Modified Date:  22.08.2022
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Any
from collections.abc import Sequence
from ampel.base.AmpelBaseModel import AmpelBaseModel
from ampel.model.purge.PurgeModel import PurgeModel
#from ampel.model.ViewModel import ViewModel


class ChannelModel(AmpelBaseModel):

	channel: int | str
	version: None | int | float | str
	#purge: PurgeModel = PurgeModel(
	#	content={'delay': 100, 'format': 'json', 'unify': True},
	#	logs={'delay': 50, 'format': 'csv'}
	#)
	# view: str = "MongoChannelView"
	active: bool = True
	hash: None | int
	distrib: None | str
	source: None | str
	contact: None | str
	access: Sequence[str] = []
	#: Identities allowed to access data associated with this channel
	members: None | Sequence[str]
	policy: Sequence[str] = []



import pytest
import pathlib
import secrets

from ampel.model.job.JobModel import JobModel
from ampel.model.job.ExpressionParser import ExpressionParser


def test_resolve_parameters():
    job = JobModel(
        **{
            "name": "job",
            "task": [
                {
                    "unit": "flerp",
                    "config": {
                        "param": "foo{{ job.parameters.param }}bar",
                        "item": "{{ item }}",
                    },
                }
            ],
            "parameters": [
                {"name": "param", "value": "-biz-"},
            ],
        }
    )
    assert job.resolve_expressions(job.task[0].dict(), job.task[0], item="scalar_item")[
        "config"
    ] == {
        "param": "foo-biz-bar",
        "item": "scalar_item",
    }


@pytest.mark.parametrize(
    ["expand_with", "result"],
    [
        ({"items": ["flim", 0, {}]}, ["flim", 0, {}]),
        ({"sequence": {"count": 1}}, ["0"]),
        ({"sequence": {"start": 1, "end": 3}}, ["1", "2"]),
    ],
)
def test_expand_with_items(expand_with, result):
    job = JobModel(
        **{
            "name": "job",
            "task": [
                {
                    "unit": "flerp",
                    "config": {
                        "item": "{{ item }}",
                    },
                    "expand_with": expand_with,
                }
            ],
        }
    )
    assert list(job.task[0].expand_with) == result


def test_multiplier_maps_to_expand():
    """
    multiplier keyword can be used as a synonym for expand_with
    """
    job = JobModel(
        **{
            "name": "job",
            "task": [
                {
                    "unit": "flerp",
                    "config": None,
                    "multiplier": 2,
                }
            ],
        }
    )
    assert list(job.task[0].expand_with) == ["0", "1"]


def test_resolve_task_outputs(tmp_path):
    token_path = tmp_path / "token"
    token_path.write_text(token := secrets.token_hex())
    job = JobModel(
        **{
            "name": "job",
            "task": [
                {
                    "unit": "bloop",
                    "config": {"output": str(token_path)},
                    "outputs": {
                        "parameters": [
                            {"name": "token", "value_from": {"path": str(token_path)}}
                        ]
                    },
                },
                {
                    "unit": "flerp",
                    "config": {"token": "{{ task.bloop.outputs.parameters.token }}"},
                },
            ],
        }
    )
    job.resolve_expressions(job.task[1].dict(), job.task[1])["config"]["token"] == token


def test_evaluate_expression():
    with pytest.raises(ValueError):
        ExpressionParser.evaluate("job.parameters.param", {})
        ExpressionParser.evaluate(
            "thing.parameters.param", {"parameters": {"param": "foo"}}
        )
        ExpressionParser.evaluate(
            "parameters.param", {"thing": {"parameters": {"param": "foo"}}}
        )
    assert (
        ExpressionParser.evaluate(
            "job.parameters.param", {"job": {"parameters": {"param": "foo"}}}
        )
        == "foo"
    )
    assert ExpressionParser.evaluate("item", {"item": "foo"}) == "foo"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/t3/include/session/T3SessionLastRunTime.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                06.01.2020
# Last Modified Date:  09.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import ClassVar
from datetime import datetime, timedelta
from ampel.mongo.query.var.events import get_last_run
from ampel.abstract.AbsT3Supplier import AbsT3Supplier
from ampel.view.T3Store import T3Store


class T3SessionLastRunTime(AbsT3Supplier[dict]):

	key: ClassVar[str] = "last_run"

	#: timedelta to search for previous run
	lookup_range: dict[str, int] = {'days': -7}

	#: timedelta to use if no previous run found
	fallback: None | dict[str, int] = None  # ex: {'days': -90}


	def supply(self, t3s: T3Store) -> dict:
		"""Add last run time (UNIX epoch) as "last_run". """

		last_run = get_last_run(
			self.context.db.get_collection('event'),
			require_success = True,
			process_name = self.event_hdlr.process_name,
			gte_time = self.lookup_range,
			timestamp = True,
		)

		if last_run is None:
			self.logger.warn(f"Event {self.event_hdlr.process_name}: last run time unavailable")
			if self.fallback:
				t = (datetime.now() + timedelta(**self.fallback)).timestamp()
				self.logger.warn(f"Fallback last run time: {t}")
				return {self.key: t}

		return {self.key: last_run}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/util/concurrent.py
# License:             BSD-3-Clause
# Author:              Jakob van Santen <jakob.van.santen@desy.de>
# Date:                07.08.2020
# Last Modified Date:  07.08.2020
# Last Modified By:    Jakob van Santen <jakob.van.santen@desy.de>

"""
Simple decorator for running a function in subprocess with asyncio, adapted
from pebble.concurrent.process. Like pebble, the future returned by the
decorated function can be cancelled to terminate the underlying subprocess.
Unlike pebble (or concurrent.futures.ProcessPoolExecutor), no extra Python
threads are needed to manage the process lifecycle.
"""

import asyncio, io, itertools, os, signal, sys, traceback
from typing import Any
from functools import wraps, partial
from multiprocessing import reduction, spawn  # type: ignore
from multiprocessing.context import set_spawning_popen
from subprocess import _args_from_interpreter_flags  # type: ignore

import ampel.vendor.aiopipe as aiopipe  # type: ignore
from ampel.metrics.prometheus import prometheus_cleanup_worker, prometheus_setup_worker


def process(function=None, **kwargs):
    """
    Runs the decorated function in a concurrent process. All arguments and
    return values must be pickleable.

    The decorated function will return an asyncio.Task that can be awaited in
    an event loop. The task will complete when the function returns or raises
    an exception. If the task is cancelled, the process will be terminated.
    """
    if function is None:
        return partial(_process_wrapper, **kwargs)
    else:
        return _process_wrapper(function)


class RemoteTraceback(Exception):
    """Traceback wrapper for exceptions in remote process.

    Exception.__cause__ requires a BaseException subclass.

    """

    def __init__(self, traceback):
        self.traceback = traceback

    def __str__(self):
        return self.traceback


class RemoteException(BaseException):
    """Pickling wrapper for exceptions in remote process."""

    def __init__(self, exception, traceback):
        self.exception = exception
        self.traceback = traceback

    def __reduce__(self):
        return rebuild_exception, (self.exception, self.traceback)


def rebuild_exception(exception, traceback):
    exception.__cause__ = RemoteTraceback(traceback)

    return exception


def prepare(data):
    "stripped-down version of multiprocessing.spawn.prepare()"
    if "sys_path" in data:
        sys.path = data["sys_path"]

    if "sys_argv" in data:
        sys.argv = data["sys_argv"]

    if "dir" in data:
        os.chdir(data["dir"])

    if "init_main_from_name" in data:
        spawn._fixup_main_from_name(data["init_main_from_name"])
    elif "init_main_from_path" in data:
        spawn._fixup_main_from_path(data["init_main_from_path"])

    if "process_name" in data:
        prometheus_setup_worker({"process": data["process_name"]})


def spawn_main(read_fd, write_fd):
    """
    Execute pickled _Process received over pipe
    """
    
    with open(read_fd, "rb") as rx:
        preparation_data = reduction.pickle.load(rx)
        prepare(preparation_data)
        obj = reduction.pickle.load(rx)
    ret = None
    exitcode = 1
    try:
        ret = obj()
        exitcode = 0
        payload = reduction.pickle.dumps(ret)
    except Exception as error:
        error.traceback = traceback.format_exc()
        ret = RemoteException(error, error.traceback)
        payload = reduction.pickle.dumps(ret)
    try:
        with open(write_fd, "wb") as tx:
            tx.write(payload)
    except Exception:
        print(f"Process {obj._name} (pid {os.getpid()}):", file=sys.stderr)
        traceback.print_exc(file=sys.stderr)
        exitcode = 1

    sys.exit(exitcode)


class _Process:
    _counter = itertools.count(1)
    #: PIDs for active processes str -> (pid -> replica), used by AmpelProcessCollector
    _active: dict[str, dict[int, int]] = {}
    #: Replica ids that can be recycled
    _expired: dict[str, set[int]] = {}

    def __init__(self, target=None, name=None, timeout=3.0, args=(), kwargs={}):
        self._target = target
        count = next(self._counter)
        # infer the process name from a ProcessModel-like dict
        if name is None:
            for arg in args:
                if isinstance(arg, dict) and "name" in arg:
                    self._name = arg["name"]
                    break
            else:
                self._name = f"{count}"
        else:
            self._name = name
        self._timeout = timeout
        self._args = tuple(args)
        self._kwargs = dict(kwargs)

    def __call__(self):
        if self._target:
            return self._target(*self._args, **self._kwargs)

    def _get_command_line(self, read_fd, write_fd):
        return (
            spawn.get_executable(),
            *_args_from_interpreter_flags(),
            "-c",
            f"from ampel.util.concurrent import spawn_main; spawn_main({read_fd}, {write_fd})",
        )

    async def launch(self) -> Any:
        prep_data = spawn.get_preparation_data(self._name)
        prep_data["process_name"] = self._name
        fp = io.BytesIO()
        set_spawning_popen(self)
        try:
            reduction.dump(prep_data, fp)
            reduction.dump(self, fp)
        finally:
            set_spawning_popen(None)

        parent_r, child_w = aiopipe.aiopipe()
        child_r, parent_w = aiopipe.aiopipe()

        with child_r.detach() as crx, child_w.detach() as ctx:
            proc = await asyncio.subprocess.create_subprocess_exec(
                *self._get_command_line(crx._fd, ctx._fd),
                pass_fds=sorted(p._fd for p in (crx, ctx)),
                start_new_session=True,
            )

        # Processes labeled with the same name are identified by replica number
        # for monitoring purposes. Numbers are reused from smallest to largest
        # in order to keep their cardinality as small as possible, e.g. if
        # replicas 0,1,3,4 are live, the next number should be 2.
        if self._name in self._active:
            replica_idx = expired.pop() if (expired := self._expired.get(self._name, None)) else max(self._active[self._name].keys())+1
            self._active[self._name][replica_idx] = proc.pid
        else:
            replica_idx = 0
            self._active[self._name] = {replica_idx: proc.pid}

        try:
            async with parent_w.open() as tx:
                tx.write(fp.getbuffer())
                await tx.drain()

            async with parent_r.open() as rx:
                try:
                    exitcode, payload = await asyncio.gather(
                        proc.wait(), rx.read(), return_exceptions=True
                    )
                    if isinstance(exitcode, BaseException):
                        raise exitcode
                    elif exitcode < 0:
                        signame = signal.Signals(-exitcode).name
                        raise RuntimeError(f"Process {self._name} (pid {proc.pid}) died on {signame}")
                    if isinstance(payload, BaseException):
                        raise payload
                    else:
                        ret = reduction.pickle.loads(payload) # type: ignore
                    if isinstance(ret, BaseException):
                        raise ret
                    else:
                        return ret
                except asyncio.CancelledError:
                    proc.terminate()
                    try:
                        await asyncio.wait_for(proc.wait(), self._timeout)
                    except asyncio.TimeoutError:
                        proc.kill()
                    await asyncio.gather(proc.wait(), rx.read())
                    raise
        finally:
            # Clean up replica numbers and pids
            del self._active[self._name][replica_idx]
            if not self._active[self._name]:
                # only process active, reset everything
                del self._active[self._name]
                self._expired.pop(self._name, None)
            else:
                # at least one other process of the same name; recycle replica
                if self._name in self._expired:
                    self._expired[self._name].add(replica_idx)
                else:
                    self._expired[self._name] = {replica_idx}
            if "PROMETHEUS_MULTIPROC_DIR" in os.environ:
                prometheus_cleanup_worker(proc.pid)


_registered_functions = {}


def _register_function(function):
    global _registered_functions

    _registered_functions[(function.__qualname__, function.__module__)] = function


def _trampoline(name, module, *args, **kwargs):
    """Trampoline function for decorators.

    Lookups the function between the registered ones;
    if not found, forces its registering and then executes it.

    """
    function = _function_lookup(name, module)

    return function(*args, **kwargs)


def _function_lookup(name, module):
    """Searches the function between the registered ones.
    If not found, it imports the module forcing its registration.

    """
    if module == "__main__":
        module = "__mp_main__"
    try:
        return _registered_functions[(name, module)]
    except KeyError:  # force function registering
        __import__(module)
        return _registered_functions[(name, module)]


def _process_wrapper(function, timeout=3.0):
    # keep the wrapped function so we can actually call it
    _register_function(function)

    @wraps(function)
    def wrapper(*args, **kwargs):
        target = _trampoline
        args = [function.__qualname__, function.__module__] + list(args)
        proc = _Process(target=target, timeout=timeout, args=args, kwargs=kwargs)
        return asyncio.create_task(proc.launch())

    return wrapper



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# File:                Ampel-core/ampel/abstract/AbsBufferComplement.py
# License:             BSD-3-Clause
# Author:              valery brinnel <firstname.lastname@gmail.com>
# Date:                16.01.2020
# Last Modified Date:  13.12.2021
# Last Modified By:    valery brinnel <firstname.lastname@gmail.com>

from typing import Iterable
from ampel.types import Traceless
from ampel.view.T3Store import T3Store
from ampel.protocol.LoggerProtocol import LoggerProtocol
from ampel.base.AmpelABC import AmpelABC
from ampel.base.decorator import abstractmethod
from ampel.struct.AmpelBuffer import AmpelBuffer
from ampel.core.ContextUnit import ContextUnit


# Inherits ContextUnit because implementing classes might need access to
# an AmpelConfig instance (foremost to the contained resource definitions)
class AbsBufferComplement(AmpelABC, ContextUnit, abstract=True):
	"""
	Complement :class:`~ampel.core.AmpelBuffer.AmpelBuffer` with information
	stored outside the Ampel database.
	"""

	logger: Traceless[LoggerProtocol]

	@abstractmethod
	def complement(self, it: Iterable[AmpelBuffer], t3s: T3Store) -> None:
		"""Add information to each :class:`~ampel.core.AmpelBuffer.AmpelBuffer` """
		...
